SystemOrganization addCategory: #SandstoneDb!
SystemOrganization addCategory: #'SandstoneDb-Tests'!

!FileDirectory methodsFor: '*sandstonedb' stamp: 'rjl 4/16/2008 12:17'!
fileNamesRecursively
	"Answer a collection containing the file names of all the files 
	in the subtree of the file system whose root is this directory."
	| result todo dir |
	result := OrderedCollection new: 100.
	todo := OrderedCollection with: self.
	[ todo notEmpty ] whileTrue: 
		[ dir := todo removeFirst.
		result addAll: dir fileNames.
		todo addAll: (dir directoryNames collect: [ :n | dir directoryNamed: n ]) ].
	^ result asArray! !

!ArrayedCollection class methodsFor: '*sandstonedb-as yet unclassified' stamp: 'sas 4/9/2008 13:25'!
deservesSandstoneReferenceMark
	"Answers true if the receiver deserves 
	to be marked (or sub marked) as Sandstone reference"
	
	^ false! !

!DataStream class methodsFor: '*sandstonedb' stamp: 'rjl 1/14/2008 08:56'!
fileNamed: aName do: aBlock 
	| file |
	file := self fileNamed: aName.
	^ file 
		ifNil: [ nil ]
		ifNotNil: [ [ aBlock value: file ] ensure: [ file close ] ]! !

!DataStream class methodsFor: '*sandstonedb' stamp: 'rjl 1/14/2008 08:57'!
newFileNamed: aName do: aBlock 
	| file |
	file := self newFileNamed: aName.
	^ file 
		ifNil: [ nil ]
		ifNotNil: [ [ aBlock value: file ] ensure: [ file close ] ]! !

!DataStream class methodsFor: '*sandstonedb' stamp: 'rjl 1/14/2008 08:57'!
oldFileNamed: aName do: aBlock 
	| file |
	file := self oldFileNamed: aName.
	^ file 
		ifNil: [ nil ]
		ifNotNil: [ [ aBlock value: file ] ensure: [ file close ] ]! !

!DataStream class methodsFor: '*sandstonedb' stamp: 'rjl 1/12/2008 16:51'!
readOnlyFileNamed: aString
	| strm |
	strm := self on: (FileStream readOnlyFileNamed: aString).
	strm byteStream setFileTypeToObject.
	^ strm! !

!DataStream class methodsFor: '*sandstonedb' stamp: 'rjl 1/14/2008 08:56'!
readOnlyFileNamed: aName do: aBlock 
	| file |
	file := self readOnlyFileNamed: aName.
	^ file 
		ifNil: [ nil ]
		ifNotNil: [ [ aBlock value: file ] ensure: [ file close ] ]! !

TestCase subclass: #SDActiveRecordTests
	instanceVariableNames: 'mom kid'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SandstoneDb-Tests'!

!SDActiveRecordTests commentStamp: 'rjl 12/19/2007 20:41' prior: 0!
Part of the reason I did this project was to force myself to learn to do more unit testing.  I'm starting to learn to like it.!

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 4/10/2008 00:17'!
flushAndReload
	SDPersonMock
		coolDown ;
		warmUp! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 4/9/2008 23:50'!
setUp
	SDPersonMock coolDown.
	SDPersonMock ensureDirectories.
	SDManMock ensureDirectories.
	SDWomanMock ensureDirectories.
	mom := SDPersonMock testPerson.
	kid := SDPersonMock testPerson! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 4/9/2008 23:50'!
tearDown
	SDPersonMock storageDirectory recursiveDelete.
	SDPersonMock coolDown ! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 3/10/2008 20:56'!
testArraySerialization
	kid save.
	mom children: {  kid  }.
	mom save.
	self flushAndReload .
	self 
		assert: mom refreshed children
		includes: kid refreshed! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 4/1/2008 12:01'!
testBigSave
	| commitTime people |
	people := (1 to: 30) collect: [:it | SDPersonMock testPerson].
	commitTime := [people do: [:each | each save]] timeToRun milliSeconds.
	people do: [:each | each delete].
	Transcript show: commitTime.
	self assert: commitTime < 3 seconds! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 3/10/2008 20:56'!
testCollectionSerialization
	kid save.
	mom children: (OrderedCollection with: kid).
	mom save.
	self flushAndReload .
	self 
		assert: mom refreshed children
		includes: kid refreshed! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 12/19/2007 21:04'!
testCreatedOn
	kid save.
	self assert: kid createdOn <= TimeStamp now! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 12/30/2007 23:56'!
testDelete
	kid save.
	self deny: kid isNew.
	self assert: kid version equals: 1.
	kid delete.
	self assert: kid isNew.
	self assert: kid version equals: 0! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 2/4/2008 10:46'!
testDeleteAndFind
	kid name: 'zorgle'.
	kid save.
	self deny: kid isNew.
	kid delete.
	self assert: (SDPersonMock find: [:e | e name = 'zorgle']) isNil.
! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 3/10/2008 20:56'!
testDictionarySerialization
	kid save.
	mom children: (Dictionary new
			at: #son
				put: kid ;
			yourself).
	mom save.
	self flushAndReload.
	self 
		assert: (mom refreshed children at: #son)
		equals: kid refreshed! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 3/10/2008 20:54'!
testEquality
	mom save.
	kid mother: mom.
	kid save.
	self flushAndReload .
	self assert: kid refreshed mother equals: mom refreshed! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 3/10/2008 20:22'!
testFind
	kid save.
	self flushAndReload.
	self deny: (SDPersonMock find: [:each | each id = kid id]) isNil.
	self assert: (SDPersonMock find: [:each | each id = 'not']) isNil! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 12/19/2007 20:17'!
testFindById
	kid save.
	self deny: (SDPersonMock findById: kid id) isNil! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 3/10/2008 20:52'!
testFindIdentity
	mom save.
	self flushAndReload .
	self 
		assert: (SDPersonMock findById: mom id) == (SDPersonMock findById: mom id)! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 4/1/2008 12:11'!
testFindSubclasses
	| man woman |
	man := SDManMock testPerson save.
	woman := SDWomanMock testPerson save.
	mom save.
	self assert: 3 equals: SDPersonMock findAll size.
	self assert: (SDPersonMock findAll 
				contains: [:e | e class = SDManMock]).
	self assert: 1 equals: SDManMock findAll size.
	self assert: 1 equals: SDWomanMock findAll size .
	man delete.
	woman delete! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 3/10/2008 20:54'!
testIdentity
	mom save.
	kid mother: mom.
	kid save.
	self flushAndReload.
	self assert: kid refreshed mother == mom refreshed! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 3/10/2008 20:55'!
testIsNew
	self assert: kid isNew.
	kid save.
	self deny: kid isNew.
	kid delete.
	self assert: kid isNew.! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 1/7/2008 17:57'!
testLoadMissingFile
	kid save.
	(kid class storageDirFor: kid id) assureExistence deleteFileNamed: kid id, '.obj'.
	self 
		should: [ kid class loadObjectFrom: kid id ]
		raise: SDLoadError! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 4/9/2008 23:46'!
testMarkReferences
	kid mother: mom.
	mom updateForSave.	"fake the save"
	kid sandstoneMarkReferences.
	self assert: (kid mother isKindOf: SDRecordMarker)! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 3/24/2008 18:04'!
testMarkReferencesCopies
	kid save.
	mom children: {kid}.
	mom save.
	self assert: mom children first == kid! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 4/9/2008 23:46'!
testMarkReferencesRecursive
	kid buddy: #not -> mom.
	mom updateForSave.	"fake the save"
	kid sandstoneMarkReferences.
	self assert: (kid buddy value isKindOf: SDRecordMarker)! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 4/9/2008 23:47'!
testMarkReferencesRecursiveDeeper
	kid buddy: #not -> (#deeper -> mom).
	mom updateForSave.	"fake the save"
	kid sandstoneMarkReferences.
	self assert: (kid buddy value value isKindOf: SDRecordMarker)! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 4/9/2008 23:47'!
testMarkReferencesRecursiveDeeperInCollection
	kid buddy: #not -> {  (#deeper -> mom)  }.
	mom updateForSave.	"fake the save"
	kid sandstoneMarkReferences.
	self assert: (kid buddy value first value isKindOf: SDRecordMarker)! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 4/9/2008 23:47'!
testMarkReferencesRecursiveDeeperInDictionary
	kid buddy: #not -> { (Dictionary new 	at: #deeper put: mom ; yourself) }.
	mom updateForSave.	"fake the save"
	kid sandstoneMarkReferences.
	self assert: ((kid buddy value first at: #deeper) isKindOf: SDRecordMarker)! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 4/9/2008 23:47'!
testMarkReferencesRecursiveDeeperNestedList
	kid buddy: #not -> (Array with: (Array with: mom)).
	mom updateForSave. "fake the save"
	kid sandstoneMarkReferences.
	self assert: (kid buddy value first first isKindOf: SDRecordMarker).
! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 4/9/2008 23:47'!
testMarkReferencesRecursiveDeeperNotTouchedInOrig
	| otherKid |
	kid buddy: #not -> (#deeper -> mom).
	otherKid := kid deepClone.
	otherKid buddy value value updateForSave. "fake the save"
	otherKid sandstoneMarkReferences.
	self assert: (kid buddy value value isKindOf: mom class)! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 3/19/2008 00:01'!
testPeerIdentity
	mom save.
	kid mother: mom.
	kid save.
	self flushAndReload .
	self assert: kid refreshed mother == mom refreshed.! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 4/10/2008 21:07'!
testResolveReferences
	mom save.
	kid mother: mom asReferenceMarker.
	kid sandstoneResolveReferences.
	self assert: (kid mother isKindOf: SDActiveRecord)! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 4/10/2008 21:07'!
testResolveReferencesRecursive
	mom save.
	kid buddy: #not -> mom asReferenceMarker.
	kid sandstoneResolveReferences.
	self assert: (kid buddy value isKindOf: SDActiveRecord)! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 4/10/2008 21:07'!
testResolveReferencesRecursiveDeeperNestedList
	mom save.
	kid buddy: #not -> (Array with: (Array with: mom asReferenceMarker)).
	kid sandstoneResolveReferences.
	self assert: (kid buddy value first first isKindOf: SDActiveRecord)! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 12/30/2007 19:08'!
testSaveMissingFile
	self assert: kid isNew.
	kid save.
	self deny: kid isNew.
	(kid class storageDirFor: kid id) assureExistence deleteFileNamed: kid id.
	kid save.
	self deny: kid isNew.! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 3/10/2008 20:57'!
testSetSerialization
	kid save.
	mom children: (Set with: kid).
	mom save.
	self flushAndReload.
	self 
		assert: mom refreshed children
		includes: kid refreshed! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 4/10/2008 20:05'!
testStorageDir
	"Active records id's must find a proper subdirectory entry in the defined structure"
	| ids legalNames |
	legalNames := (0 to: 9) collect: [ : e | e asString ].
	ids := Set new.
	1000 timesRepeat: [ ids add: UUID new asString36 ].
	ids add: 'abaoblwgnaydxokccorveamoq'.
	ids do: [ : anId | self assert: (legalNames includes: (SDPersonMock storageDirFor: anId) localName) ]! !

!SDActiveRecordTests methodsFor: 'running' stamp: 'rjl 12/19/2007 21:07'!
testUpdatedOn
	kid save.
	self assert: kid updatedOn <= TimeStamp now! !

!ExternalType class methodsFor: '*sandstonedb-as yet unclassified' stamp: 'sas 4/9/2008 13:30'!
deservesSandstoneReferenceMark
	"Answers true if the receiver deserves 
	to be marked (or sub marked) as Sandstone reference"
	
	^ false! !

!Object class methodsFor: '*sandstonedb-testing' stamp: 'sas 4/9/2008 13:11'!
deservesSandstoneReferenceMark
	"Answers true if the receiver deserves 
	to be marked (or sub marked) as Sandstone reference"
	
	^ true! !

!Object class methodsFor: '*sandstonedb' stamp: 'rjl 4/10/2008 20:22'!
sandstoneReferenceMarkerSupport
	"Answers a reference marker support which 
	is specialized in the instropection of the receiver"
	^ SDReferenceMarker! !

!Object class methodsFor: '*sandstonedb' stamp: 'rjl 4/10/2008 21:07'!
sandstoneReferenceResolverSupport
	"Answers a support to reify the receiver
	references to sandstone active records"
	^ SDReferenceResolver! !

!Object methodsFor: '*sandstonedb-testing' stamp: 'sas 4/9/2008 12:53'!
deservesSandstoneReferenceMark
	"Answers true if the receiver deserves 
	to be marked (or sub marked) as Sandstone reference"

	^ self class deservesSandstoneReferenceMark! !

!Object methodsFor: '*sandstonedb-testing' stamp: 'sas 4/9/2008 13:20'!
isSandstoneActiveRecord
	"Answers true if the receiver is
	a Sandstone Active Record."

	^ false! !

!Object methodsFor: '*sandstonedb-serialization' stamp: 'rjl 4/10/2008 21:21'!
sandstoneMarkReferences
	"Replaces the receiver (sub) active records 
	with Sandstone references where it is needed"

	^ self sandstoneReferenceMarker markReferences! !

!Object methodsFor: '*sandstonedb-serialization' stamp: 'rjl 4/10/2008 21:25'!
sandstoneMarkReferencesVisits: visitedParts 
	"Replaces the receiver (sub) active records 
	with Sandstone references"
	
	^ self sandstoneReferenceMarker markReferencesVisits: visitedParts! !

!Object methodsFor: '*sandstonedb-factory' stamp: 'rjl 4/10/2008 21:19'!
sandstoneReferenceMarker
	"Answers a reference marker which 
	is specialized in the instropection of the receiver"
	
	^ self class sandstoneReferenceMarkerSupport for: self! !

!Object methodsFor: '*sandstonedb-factory' stamp: 'rjl 4/10/2008 21:20'!
sandstoneReferenceResolver
	"Answers a reference reifier which 
	is specialized in the instropection of the receiver"
	
	^ self class sandstoneReferenceResolverSupport for: self! !

!Object methodsFor: '*sandstonedb-serialization' stamp: 'rjl 4/10/2008 21:22'!
sandstoneResolveReferences
	"Replaces the receiver markers with
	active records where it is needed"
	
	^ self sandstoneReferenceResolver resolveReferences! !

!Object methodsFor: '*sandstonedb-serialization' stamp: 'rjl 4/10/2008 21:25'!
sandstoneResolveReferencesVisits: visitedParts 
	"Replaces the receiver (sub) active records 
	with Sandstone references"
	
	^ self sandstoneReferenceResolver resolveReferencesVisits: visitedParts! !

Object subclass: #SDActiveRecord
	instanceVariableNames: 'id createdOn updatedOn version isNew'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SandstoneDb'!

SDActiveRecord class
	instanceVariableNames: 'locks cache lock'!

!SDActiveRecord commentStamp: 'rjl 4/1/2008 12:26' prior: 0!
I'm the core of a simple object database based upon a mix of several patterns intended for use with *small* Seaside based web applications.  I am not meant to scale to millions of records, but for prototyping and small office applications where the number of records are in the thousands and the number of concurrent users can be handled by a single Squeak image.  To use me, simply subclass me, then call YouClass install, that's it.  The database is kept in a subdirectory matching the name of the class in the same directory as the image.  This is a prevayler like system, all data is kept in memory and logged to disk, on system startup, all data is loaded from disk into memory.  The keeps the image small and read only.

ActiveRecord
Simple api, #do, #find, #findAll on the class and #save and #delete on each instance with no configuration *at all* beyond the initial #install and completely avoids the write barrier issue by making the programmer responsible for saving changes rather than tracking and committing changed objects.  Tracking and commiting changed objects is what makes GOODS and Magma so slow with large commits because they have to check every object in the cache for changes even if only one actually changes.  They're also both server based which makes things like sessions and connection pools necessary, much more complex than most web apps need.  Every record contains some basic fields, id, createdOn, updatedOn, version, and isNew. 

Domain Aggregates
Each record is a root aggregate, child objects are serialized in one chunk with their parent aggregate and can not be queried or referenced outside the context of the root aggregate.  Ideas taken from Eric Evans' Domain Driven Design.  This is good design, and keeps the domain clean and the biz code where it belongs.  Relational databases suck at doing this.  Basically this is a document oriented database where you stored structured objects in rather large usable chunks such as an order that contains all it's items or an auction that contains all it's bids.

Rather than transactions across many statements or tables, you just save documents (complex objects) where each save either succeeds or fails.

Here are some trivial examples..

person := Person find: [ :e | e name = 'Joe' ].
managers := Employee findAll: [ :e | e hasSubordinates ].
person save.
person delete.

"commit is a critical section with an implicit save at the end"
auction commit: [
	auction addBid: Bid price: 30 dollars.
]

"critical is an object level lock, you must call save yourself"
auction critical: [
	[auction addBid: (Bid price: 30 dollars).
	auction save.
	auction sendPendingEmails] 
	on: Error do:[ :error | 
		auction flushPendingEmails.
		error asDebugEmail]]
!

SDActiveRecord class
	instanceVariableNames: 'locks cache lock'!

!SDActiveRecord class methodsFor: 'private' stamp: 'rjl 3/31/2008 14:19'!
allCaches
	^ self allSubclasses 
		inject: self cache
		into: [ :sum :subclass | sum , subclass cache ]! !

!SDActiveRecord class methodsFor: 'private' stamp: 'rjl 3/31/2008 14:01'!
allCurrentVersionNames
	"Recursively searching for files allows the directory paths to be hashed to split the files up into many directories to break the max file per directoy limit and keep things fast, and makes queries return all instances of this class and it's subclasses, a nice simple way to deal with inheritance"
	^ self storageDirectory fileNamesRecursively select: [ :each | each endsWith: '.obj' ]! !

!SDActiveRecord class methodsFor: 'private' stamp: 'rjl 2/4/2008 18:00'!
cache
	^ cache ifNil: [ cache := Dictionary new ]! !

!SDActiveRecord class methodsFor: 'private' stamp: 'rjl 2/4/2008 17:46'!
cache: aCache
	cache := aCache! !

!SDActiveRecord class methodsFor: 'private' stamp: 'rjl 4/10/2008 21:07'!
cachedLoadFrom: anId 
	"seems I have to make sure to cache the object before I can resolve it's references so any backreferences to it don't try and load from disk again''"
	(self cache includesKey: anId) ifFalse: 
		[ ([ self loadObjectFrom: anId ] 
			on: SDLoadError
			do: [ nil ]) ifNotNilDo: 
			[ : it | 
			self cache 
				at: anId
				put: it.
			it sandstoneResolveReferences ] ].
	^ self cache 
		at: anId
		ifAbsent: [ nil ]! !

!SDActiveRecord class methodsFor: 'actions' stamp: 'rjl 4/9/2008 23:50'!
coolDown
	cache := nil.
	self allSubclassesDo: [ :e | e coolDown ]! !

!SDActiveRecord class methodsFor: 'accessing' stamp: 'rjl 4/1/2008 12:14'!
defaultBaseDirectory
	"you can override this if you want to force the db somewhere else"
	^ FileDirectory default! !

!SDActiveRecord class methodsFor: 'crash recovery' stamp: 'rjl 4/2/2008 15:06'!
deleteFailedCommits
	"all remaining .new files are failed commits, kill them"
	[(self storageDirectory fileNamesRecursively
		select: [:each | each endsWith: '.new'])
		do: [:each | (self storageDirFor: each)
				deleteFileNamed: each]]
		on: Error
		do: [:err | Transcript show: err]! !

!SDActiveRecord class methodsFor: 'private' stamp: 'rjl 4/10/2008 20:07'!
dirNameFor: anId 
	"Answers a string with one decimal digit 
	corresponding to anId."
	^ (anId 
		inject: 0
		into: [ : sum : e | sum + e asInteger ]) asReducedSumOfDigits asString! !

!SDActiveRecord class methodsFor: 'queries' stamp: 'rjl 3/31/2008 14:18'!
do: aBlock
	"do on a copy in case the do modifies the collection I'm trying to iterate'"
	self findAll copy do: aBlock! !

!SDActiveRecord class methodsFor: 'private' stamp: 'rjl 3/31/2008 14:07'!
ensureDirectories
	| dir |
	self ~= SDActiveRecord ifTrue: 
		[ dir := self storageDirectory assureExistence.
		(dir directoryNamed: '0') assureExistence.
		(dir directoryNamed: '1') assureExistence.
		(dir directoryNamed: '2') assureExistence.
		(dir directoryNamed: '3') assureExistence.
		(dir directoryNamed: '4') assureExistence.
		(dir directoryNamed: '5') assureExistence.
		(dir directoryNamed: '6') assureExistence.
		(dir directoryNamed: '7') assureExistence.
		(dir directoryNamed: '8') assureExistence.
		(dir directoryNamed: '9') assureExistence ].
	self allSubclassesDo: [ :e | e ensureDirectories ]! !

!SDActiveRecord class methodsFor: 'queries' stamp: 'rjl 3/31/2008 14:19'!
find: aBlock 
	^ self 
		find: aBlock
		ifAbsent: [ nil ]! !

!SDActiveRecord class methodsFor: 'queries' stamp: 'rjl 3/31/2008 14:20'!
find: aBlock ifAbsent: aHandler 
	^ self findAll 
		detect: aBlock
		ifNone: aHandler! !

!SDActiveRecord class methodsFor: 'queries' stamp: 'rjl 3/31/2008 14:20'!
find: aBlock ifPresent: aHandler 
	^ self findAll 
		detect: aBlock
		ifFound: aHandler! !

!SDActiveRecord class methodsFor: 'queries' stamp: 'rjl 2/4/2008 18:04'!
findAll
	^ self allCaches values! !

!SDActiveRecord class methodsFor: 'queries' stamp: 'rjl 3/31/2008 14:20'!
findAll: aBlock 
	^ self findAll select: aBlock! !

!SDActiveRecord class methodsFor: 'queries' stamp: 'rjl 3/31/2008 14:20'!
findById: anId 
	^ self 
		findById: anId
		ifAbsent: [ nil ]! !

!SDActiveRecord class methodsFor: 'queries' stamp: 'rjl 3/31/2008 14:20'!
findById: anId ifAbsent: aHandler 
	^ self cache 
		at: anId
		ifAbsent: aHandler! !

!SDActiveRecord class methodsFor: 'crash recovery' stamp: 'rjl 4/2/2008 14:40'!
finishPartialCommits
	"find where .new exists but .obj doesn't, rename .new to
	.obj to finish commit'"
	[(self storageDirectory fileNamesRecursively
		select: [:each | each endsWith: '.new'])
		do: [:each | 
			| objFile | 
			objFile := each copyReplaceAll: '.new' with: '.obj'.
			((self storageDirFor: each)
					fileExists: objFile)
				ifFalse: [(self storageDirFor: each)
						rename: each
						toBe: objFile]]]
		on: Error
		do: [:err | Transcript show: err]! !

!SDActiveRecord class methodsFor: 'config' stamp: 'rjl 4/1/2008 09:51'!
install
	"call the lock just to initialize it"
	self lock.
	self ensureDirectories.
	Smalltalk addToStartUpList: self.
	Smalltalk addToShutDownList: self.
	self allSubclassesDo: [ :e | e install ]! !

!SDActiveRecord class methodsFor: 'private' stamp: 'rjl 4/2/2008 15:04'!
loadObjectFrom: anId 
	| objFile |
	objFile := anId , '.obj'.
	^ [SmartRefStream
		readOnlyFileNamed: ((self storageDirFor: anId)
				fullNameFor: objFile)
		do: [:file | file sandstoneNextAndClose]]
		on: Error
		do: [:err | 
			Transcript show: err;
				 show: ' - ' , objFile;
				 cr.
			SDLoadError signal: err messageText]! !

!SDActiveRecord class methodsFor: 'private' stamp: 'rjl 3/27/2008 13:16'!
lock
	^ lock ifNil: [ lock := Semaphore forMutualExclusion ]! !

!SDActiveRecord class methodsFor: 'private' stamp: 'rjl 3/31/2008 14:21'!
lockFor: anInstance 
	"keeping a weak dictionary of per instance locks for commit concurrency at the application level"
	locks ifNil: [ locks := WeakKeyDictionary new ].
	^ self lock critical: 
		[ locks 
			at: anInstance
			ifAbsentPut: [ Semaphore forMutualExclusion ] ]! !

!SDActiveRecord class methodsFor: 'system startup' stamp: 'rjl 4/9/2008 23:51'!
shutDown
	"keep the image really small, database will be reloaded on startup"
	self coolDown! !

!SDActiveRecord class methodsFor: 'system startup' stamp: 'rjl 4/10/2008 00:17'!
startUp
	"clear and then reload entire db into ram on a background thread"
	self = SDActiveRecord ifTrue: [ ^ self ].
	
	[ self ensureDirectories.
	self coolDown.
	self warmUp ] 
		forkAt: Processor systemBackgroundPriority
		named: 'loading ' , self name! !

!SDActiveRecord class methodsFor: 'private' stamp: 'rjl 4/10/2008 20:08'!
storageDirFor: anId 
	"Grab the correct hashed subdirectory for this record"
	^ self storageDirectory directoryNamed: (self dirNameFor: anId)! !

!SDActiveRecord class methodsFor: 'private' stamp: 'rjl 3/31/2008 14:07'!
storageDirectory
	"compute the path of superclasses all the way up to ActiveRecord, storing subclass records as a subdirectory of the superclasses directory allows ActiveRecord to deal with inheritance"
	| dir parentClass |
	dir := self defaultBaseDirectory.
	parentClass := self superclass.
	[ parentClass == SDActiveRecord ] whileFalse: 
		[ dir := dir directoryNamed: parentClass name.
		parentClass := parentClass superclass ].
	^ dir directoryNamed: self name! !

!SDActiveRecord class methodsFor: 'actions' stamp: 'rjl 4/10/2008 00:17'!
warmUp
	| loadTime |
	loadTime := 
	[ self finishPartialCommits.
	self deleteFailedCommits.
	self allCurrentVersionNames do: 
		[ : each | 
		[ self cachedLoadFrom: (each copyUpTo: $.) ] 
			on: Error
			do: [ : err | Transcript show: err ] ] ] timeToRun milliSeconds.
	Transcript
		show: self name , ' loaded in ' , loadTime asString ;
		cr! !

!SDActiveRecord methodsFor: 'comparing' stamp: 'rjl 4/11/2008 11:39'!
= anObject 
	"asking the object isMemberOf ensures that if it's a proxy that message 
	will be forwarded to the real object.  Checking this condition in reverse
	anObject class, will fail because anObject class will be the ProxyClass"
	
	^ (anObject isMemberOf: self class) and: [ id = anObject id ]! !

!SDActiveRecord methodsFor: 'converting' stamp: 'rjl 4/10/2008 20:20'!
asReferenceMarker
	isNew ifTrue: 
		[ "*********** USE THIS EXCEPTION AS DESIGN REFLECTION INVOCATOR ***********
		Programmers may reach this point several times until they get the intended use of
		this persistency solution. This is expected **by design** to make atomic saves really small 
		and consistent. So commits of active records which have other active records as parts (at any deep)
		are restricted intentionally. They only are allowed when all its sub active records are previously commited.
		To be conscious about what consequences will have to make big objects to be atomically saved refer to:
			http://www.ccs.neu.edu/groups/IEEE/ind-acad/brewer/     and 
			http://www.enterpriseintegrationpatterns.com/ramblings/18_starbucks.html
		
		Only the programmer knows the proper logical commit order for his data!! This is not a relational database,
		if you absolutely need several objects to be atomically saved, then you should make them all part of a 
		single aggregate, all of your objects should not be active records, only your aggregate roots are active records...
			http://domaindrivendesign.org/discussion/messageboardarchive/Aggregates.html"
		SDCommitError signal: 'An object is being saved while referencing an unsaved peer of type ' , self class name , '.  You must save that record first!!' ].
	^ SDRecordMarker new
		id: id ;
		className: self class name ;
		yourself! !

!SDActiveRecord methodsFor: 'actions' stamp: 'rjl 4/16/2008 15:35'!
commit
	"convenience method for a locked save"
	self commit: []! !

!SDActiveRecord methodsFor: 'actions' stamp: 'rjl 3/27/2008 12:52'!
commit: aBlock 
	(self class lockFor: self) critical: 
		[ aBlock value.
		self save ]! !

!SDActiveRecord methodsFor: 'accessing' stamp: 'rjl 1/16/2008 18:39'!
createdOn
	^createdOn! !

!SDActiveRecord methodsFor: 'actions' stamp: 'rjl 3/27/2008 13:23'!
critical: aBlock 
	^ (self class lockFor: self) critical: aBlock! !

!SDActiveRecord methodsFor: 'copying' stamp: 'rjl 3/31/2008 12:00'!
deepClone
	| saveStream |
	saveStream := RWBinaryOrTextStream on: String new.
	(ReferenceStream on: saveStream) nextPut: self.
	saveStream position: 0.
	^(ReferenceStream on: saveStream) next! !

!SDActiveRecord methodsFor: 'actions' stamp: 'rjl 3/31/2008 14:04'!
delete
	| dir |
	dir := (self class storageDirFor: id) assureExistence.
	dir deleteFileNamed: id , '.obj'.
	self class cache removeKey: id ifAbsent: [  ].
	self initialize! !

!SDActiveRecord methodsFor: 'comparing' stamp: 'rjl 3/31/2008 14:08'!
hash
	^ id hash! !

!SDActiveRecord methodsFor: 'accessing' stamp: 'rjl 1/16/2008 18:40'!
id
	^ id! !

!SDActiveRecord methodsFor: 'accessing' stamp: 'rjl 4/16/2008 12:20'!
indexString
	^ String streamContents: 
		[ :s | 
		self allInstVarNames do: 
			[ :each | 
			(self instVarNamed: each) ifNotNilDo: 
				[ :value | 
				s
					nextPutAll: value asString;
					nextPutAll: '~~' ] ] ]! !

!SDActiveRecord methodsFor: 'initialize-release' stamp: 'rjl 1/16/2008 18:41'!
initialize
	super initialize.
	id := UUID new asString36.
	createdOn := updatedOn := TimeStamp now.
	version := 0.
	isNew := true! !

!SDActiveRecord methodsFor: 'testing' stamp: 'rjl 1/16/2008 18:40'!
isNew
	^ isNew ifNil: [ isNew := true ]! !

!SDActiveRecord methodsFor: 'testing' stamp: 'sas 4/9/2008 13:20'!
isSandstoneActiveRecord
	"Answers true if the receiver is
	a Sandstone Active Record."

	^ true! !

!SDActiveRecord methodsFor: 'actions' stamp: 'rjl 4/9/2008 23:45'!
save
	self validate.
	self class cache 
		at: id
		put: self.
	self updateForSave.
	self deepClone
		sandstoneMarkReferences ;
		saveToFile! !

!SDActiveRecord methodsFor: 'private' stamp: 'rjl 4/7/2008 11:32'!
saveToFile
	"The basic idea here is to make a save as atomic as possible and no data is corrupted, no partial writes, thus renaming files as atomic.
	Also keeping around one old version of the object just in case I want to recover the last good version, for the time being, until I trust this"
	| currentVersion newVersion oldVersion lastOldVersion dir |
	dir := self class storageDirFor: id.
	currentVersion := dir fullNameFor: id , '.obj'.
	newVersion := currentVersion , '.new'.
	oldVersion := currentVersion , '.' , (version - 1) asString.
	lastOldVersion := currentVersion , '.' , (version - 2 max: 0) asString.
	"the flush ensures all data is actually written to disk before moving on"
	SmartRefStream 
		newFileNamed: newVersion
		do: [ :file | file sandstoneNextPutObjOnly: self ; flush ].
	"on first save there won't be a current version, and fileExists is too expensive to check when it'll always be there except on first save"
	[ dir rename: currentVersion toBe: oldVersion ] ifError: ["just ignore it"].
	dir rename: newVersion 	toBe: currentVersion.
	dir deleteFileNamed: lastOldVersion
! !

!SDActiveRecord methodsFor: 'private' stamp: 'rjl 1/16/2008 18:58'!
updateForSave
	updatedOn := TimeStamp now.
	version := version + 1.
	isNew := false! !

!SDActiveRecord methodsFor: 'accessing' stamp: 'rjl 3/31/2008 14:17'!
updatedOn
	^ updatedOn! !

!SDActiveRecord methodsFor: 'actions' stamp: 'rjl 3/31/2008 14:17'!
validate
	"for subclasses to override and throw exceptions to prevent saves"! !

!SDActiveRecord methodsFor: 'accessing' stamp: 'rjl 3/31/2008 14:17'!
version
	^ version! !

SDActiveRecord subclass: #SDPersonMock
	instanceVariableNames: 'name dateOfBirth description father mother children buddy'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SandstoneDb-Tests'!

SDPersonMock subclass: #SDManMock
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SandstoneDb-Tests'!

!SDPersonMock class methodsFor: 'as yet unclassified' stamp: 'rjl 4/1/2008 11:53'!
testPerson
	^ (self new)
		name: 'mzqfmmv zptnhli';
		dateOfBirth: Date today - 12000 atRandom days;
		description: 'Rxvalpv tdpqkvv ikjcscw bigjmeb vukntxj qxshdtt wkczzio bqpqacu whluhqv.  Lqlwajh!!  Qknwpfd peeskqc oarthtv pfblwjj yndlxks.  .  Ucnlocq iuiiluk txkkzmh dshkhmq uokcwqy oiktkxl awxmpep rkkcmis xcgyoeg jodtjaf ntwijzp?  Bqpjsiq qthzwtx xcnqbag ausurqp xatoqar jytguon?  Bpigdij jgijxep bjgttpr kklsfaw rdoikwz!!  Hzdbjxc?  Ehwflyy qfcqntk mlhcjql ecbtrtk egxldpk rypalvw ndngbdl jhhdxts zrfjtaz gccrnni dwuqxwm nmifzcq sxkgkxh.  Kwtzsjv ghvjqqm lfgibdt rufcowp kavabmi bpveqsr shjzsft xqaxivy fjqydns ryyggif.

Jshooyy jsyojlv miusnaj onpdiss hypkzzd qaunpqd rrvgnws ekiswyv?  Glhxkkk?  Pzyjuot raaefbv cnxmbsd?  Jabaniw uirjkcm jocxnhd!!  Erwblox qikxfez.

Ezswaga cnrzhqi jrdikos kkcncxd eoewflm ylzilve viumpqm uiyqhvv azrezhc jjeesfd rfdgsgg wfmxfye iisisju vhjourb.

Yicqnqd zhaioos npnjjsc hxfwlfr ozynyjs!!  Qavvubu uodeedp?  Fjdgbxu ryswazg ifumqpv jtlifnd?  Blciucl kypalot?  Dstnwqj kwzpxrk bgewcyp triddbr acvrkey pwgcleb sqtxajv svzpnwk unnlmtg oqnbpsq svtnxdo!!

Ywodawk uikhwek kyqivqj iojkwiw.  Bjcsnpu jtfpmqd!!  Crwlngk qeeeuci nnakbai hncfdkk kbhehju ttdsdcf zqulfpj pmvtmfa xrusxuy wbamfee opzjdia hcrfdnz tiasrqz wkwidbg?  Tzlwcst twaeuwm tfhreal igamlby saekozk tjnxohl ogcdhva fckbaii utqvhjg?  Nrynpjl dniycpm kkvytuu bnuxrev zrvbcph.  Awfbhgg xtyffnz lwjkhdo evmvogt vqfqppp chdtxcr ktiwujg vcgqoya thgkhac ncagoxy unuukan gntyowz obmiwmf!!  Okwjujz tfghbxi jupstni xbzpgau zlxeblm!!  Llxuwqc gupeurb ltwyzzj xxanyln qgrtigr?

Vudeenm hnfwxay rtaacau ymmyxbi lpcbamg ifopjuw guhctxx radytrh yubjcjc lnjucta qinzmlf.  Cnunlvp gdhcgrq oxeojsd?  Bbkjtne kzlvdso xtmgqhj pfyroxl brmqhkt gvqnftz aupxtsg.

Krzxaye!!  Clhbjoq ubxtlsc jzhggvl updbzxe.

Vhtetst ruwpukn jdttpba?  Ctwactb ljoeiqa pubapwc cioporv?

Talkecn dbddhuw rxfybrf iwdxfdo spdbdyk vhxbgrm elqngon dytngbm knqxaqv zrltvgh snjkzig rzypaly ekabqcg.  Xxxmegt otxzkhh vxpyjkn pnhgpfx qsnrhjo ziftvuu ivvfacz nhdajef ezksurr gvfepxx.  Lqdldbo wegkzlk qxgqpux hfbtydd hbqfhea gxuksjm hvunuwa?  Lxqkddp bccukhy odljtnp nouapus dcqpkqz lgpxpcs fmaehsc cgybese gujcbqz icjzamy tvmvwlf dvtkpng!!  !!  Okqxkhh dfuulea?  Ratgmpz?  Pfmgfbk qhggllc.  Balobww?

Wqoyfli xlxenvi ngtdshb bhpmyhd.  Kxsidry wijdkem oyuutvw msunbhq vaektbt yfapmwl yjrwmxi xgzfpbk hxcmydy jgaybjh magjtcf.  Koxatqi ivvultq yptadcp ygqutkn dizybis nvsvfhx tdhcqfy ihvqsvs paakztf uzokxhk xlcfhbr wejdutf wywzrqs tnlswyo?  Glawqav nvndvdf cbqzhce iuygwer twyxquv pjklwnv tbnazjw vuvxvgi?

Rpyaiqq rnvfyzk fgbsbjc zhftunh uvmxaov vgvwzwq fpuurxc!!  Zemojud eadzsas wzfdcuv vxvulbc ahtiijq yclzbvf rspxhrc lymmwja!!  Mfmgrrb!!  Cyfuuvp iabqntx volbiij ebjihqw ypxvukv wnghvwo fhqwyuh tlmeyrl oxcmotx uemizkc xvbgzqh xsdqtcj xfbzrgg?  Jmtauxu!!  Kbigzux slwjszd manystg usxbiya kfgdygh wrclmih chqtyew orkrazt pekdcrh oexrivu bnbrerh?

Xbkcyvl xivutzr iitbwnm eouuiux iwaansa oxzygwa sosqxvi jkinurb bedwzss?  Uspwzru.  Eogmjki raowczb!!  ?  Lfzhatu ldnvhfn rzfqfcl yzhbxpo evahsye eibcwlx ygknqms terepdi acvezip lyqydat bksesji.  Sncraae sropzde xdhuuuf?  ?

Vtcevnw dqwwzqe kqmakan wboltvp axuhsio eddpqvk hmaqsst kbqyeqg qjnlkph zjrrcdr glyqfak!!  Ciofqeq qzfzdxn aumczcs yxzdrqu hxnlhmo ihnowav nqggjoi dfdxzqr.

Wiedtvt jpikavx hkxvzqi eyobbcb hxhktut qpgasux bnwnzhp xaceikr comlnmg jkdzrhn hctccgm zcjgqbc llklclt!!

Bgyypck fvgaauj bzetcjo rqixcnx ggwuzax cnknyiv vbtnkxj zkmenql ynxyqpp vqnvihq qydzixk jvsbxhw qnbphdl uddayfe!!  Glalnun gbtcspy kxnzokz ecdpgpr kcmvcdf ngczhkp oovhcik yvlzbgl?';
		yourself! !

!SDPersonMock methodsFor: 'accessing' stamp: 'rjl 3/10/2008 21:44'!
buddy
	^ buddy! !

!SDPersonMock methodsFor: 'accessing' stamp: 'rjl 3/10/2008 21:44'!
buddy: anObject
	buddy := anObject! !

!SDPersonMock methodsFor: 'accessing' stamp: 'rjl 12/11/2007 10:33'!
children
	^ children! !

!SDPersonMock methodsFor: 'accessing' stamp: 'rjl 12/11/2007 10:33'!
children: anObject
	children := anObject! !

!SDPersonMock methodsFor: 'accessing' stamp: 'rjl 12/7/2007 17:07'!
dateOfBirth
	^ dateOfBirth! !

!SDPersonMock methodsFor: 'accessing' stamp: 'rjl 12/7/2007 17:07'!
dateOfBirth: anObject
	dateOfBirth := anObject! !

!SDPersonMock methodsFor: 'accessing' stamp: 'rjl 12/7/2007 17:08'!
description
	^ description! !

!SDPersonMock methodsFor: 'accessing' stamp: 'rjl 12/7/2007 17:07'!
description: anObject
	description := anObject! !

!SDPersonMock methodsFor: 'accessing' stamp: 'rjl 12/7/2007 17:07'!
father
	^ father! !

!SDPersonMock methodsFor: 'accessing' stamp: 'rjl 12/7/2007 17:07'!
father: anObject
	father := anObject! !

!SDPersonMock methodsFor: 'accessing' stamp: 'rjl 12/7/2007 17:07'!
mother
	^ mother! !

!SDPersonMock methodsFor: 'accessing' stamp: 'rjl 12/7/2007 17:07'!
mother: anObject
	mother := anObject! !

!SDPersonMock methodsFor: 'accessing' stamp: 'rjl 12/7/2007 17:07'!
name
	^ name! !

!SDPersonMock methodsFor: 'accessing' stamp: 'rjl 12/7/2007 17:07'!
name: anObject
	name := anObject! !

!SDPersonMock methodsFor: 'accessing' stamp: 'rjl 3/10/2008 20:18'!
refreshed
	^self class findById: id! !

SDPersonMock subclass: #SDWomanMock
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SandstoneDb-Tests'!

Object subclass: #SDIntrospector
	instanceVariableNames: 'introspectee'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SandstoneDb'!

!SDIntrospector class methodsFor: 'instance creation' stamp: 'sas 4/9/2008 19:11'!
for: anObject

	^ self new initializeOn: anObject! !

!SDIntrospector methodsFor: 'initializing' stamp: 'sas 4/9/2008 17:50'!
initializeOn: anObject

	introspectee := anObject ! !

SDIntrospector subclass: #SDReferenceMarker
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SandstoneDb'!

!SDReferenceMarker commentStamp: 'rjl 4/10/2008 21:35' prior: 0!
I'm run over a graph of objects to peel out a slice containing just one active record and it's non active record children.  Other active records are considered siblings and are saved to their own files.  This keeps the commits small without the need for any kind of command logging like prevayler.!

SDReferenceMarker subclass: #SDCollectionReferenceMarker
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SandstoneDb'!

!SDCollectionReferenceMarker methodsFor: 'actions' stamp: 'rjl 4/10/2008 21:23'!
markReferencesVisits: visitedParts 
	"Make components of the introspectee which are 
	Sandstone active records to become references.
	Do this deeply (sub components)."
	introspectee class isMeta ifTrue: [ ^ introspectee ].
	introspectee deservesSandstoneReferenceMark ifFalse: [ ^ introspectee ].
	(visitedParts includes: introspectee) ifTrue: [ ^ introspectee ].
	visitedParts add: introspectee.
	introspectee do: 
		[ : anElement | 
		anElement sandstoneMarkReferencesVisits: visitedParts.
		anElement isSandstoneActiveRecord ifTrue: 
			[ introspectee 
				replaceAll: anElement
				with: anElement asReferenceMarker ] ].
	^ introspectee! !

!SDReferenceMarker class methodsFor: 'testing' stamp: 'sas 4/9/2008 13:33'!
deservesSandstoneReferenceMark
	"Answers true if the receiver deserves 
	to be marked (or sub marked) as Sandstone reference"
	
	^ false! !

!SDReferenceMarker methodsFor: 'actions' stamp: 'sas 4/9/2008 12:50'!
markReferences
	"Make components of the introspectee which are 
	Sandstone active records to become references.
	Do this deeply (sub components)."
	
	^ self markReferencesVisits: Set new! !

!SDReferenceMarker methodsFor: 'actions' stamp: 'rjl 4/10/2008 21:23'!
markReferencesVisits: visitedParts 
	"Make components of the introspectee which are 
	Sandstone active records to become references.
	Do this deeply (sub components)."
	introspectee class isMeta ifTrue: [ ^ introspectee ].
	introspectee deservesSandstoneReferenceMark ifFalse: [ ^ introspectee ].
	(visitedParts includes: introspectee) ifTrue: [ ^ introspectee ].
	visitedParts add: introspectee.
	introspectee class allInstVarNames do: 
		[ : name | 
		| var |
		var := introspectee instVarNamed: name.
		var deservesSandstoneReferenceMark ifTrue: [ var sandstoneMarkReferencesVisits: visitedParts ].
		var isSandstoneActiveRecord ifTrue: 
			[ introspectee 
				instVarNamed: name
				put: var asReferenceMarker ] ].
	^ introspectee! !

SDIntrospector subclass: #SDReferenceResolver
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SandstoneDb'!

!SDReferenceResolver commentStamp: 'rjl 4/10/2008 21:36' prior: 0!
I'm an introspector usefor for the reifying process of Sandstone active records.
When they are deserialized, could happen they have marks of other active records. I help to transform those marks in the actual objects thus reconstituting the original object graph.!

SDReferenceResolver subclass: #SDCollectionReferenceResolver
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SandstoneDb'!

!SDCollectionReferenceResolver methodsFor: 'actions' stamp: 'rjl 4/10/2008 21:23'!
resolveReferencesVisits: visitedParts 
	"Make components of the introspectee which are 
	Sandstone references to active record to become active records.
	Do this deeply (sub components)."
	introspectee class isMeta ifTrue: [ ^ introspectee ].
	(visitedParts includes: introspectee) ifTrue: [ ^ introspectee ].
	visitedParts add: introspectee.
	introspectee do: 
		[ : anElement | 
		(anElement isKindOf: SDRecordMarker) 
			ifTrue: 
				[ introspectee 
					replaceAll: anElement
					with: anElement asOriginalObject ]
			ifFalse: [ anElement sandstoneResolveReferencesVisits: visitedParts ] ].
	^ introspectee! !

!SDReferenceResolver methodsFor: 'actions' stamp: 'rjl 4/10/2008 21:03'!
resolveReferences
	"Make components of the introspectee which are 
	Sandstone references to active record to become active records.
	Do this deeply (sub components)."
	^ self resolveReferencesVisits: Set new! !

!SDReferenceResolver methodsFor: 'actions' stamp: 'rjl 4/10/2008 21:23'!
resolveReferencesVisits: visitedParts 
	"Make components of the introspectee which are 
	Sandstone references to active record to become active records.
	Do this deeply (sub components)."
	introspectee class isMeta ifTrue: [ ^ introspectee ].
	introspectee deservesSandstoneReferenceMark ifFalse: [ ^ introspectee ].
	(visitedParts includes: introspectee) ifTrue: [ ^ introspectee ].
	visitedParts add: introspectee.
	introspectee class allInstVarNames do: 
		[ : name | 
		| var |
		var := introspectee instVarNamed: name.
		var deservesSandstoneReferenceMark ifTrue: [ var sandstoneResolveReferencesVisits: visitedParts ].
		(var isKindOf: SDRecordMarker) ifTrue: 
			[ introspectee 
				instVarNamed: name
				put: var asOriginalObject ] ].
	^ introspectee! !

Object subclass: #SDRecordMarker
	instanceVariableNames: 'id className'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SandstoneDb'!

!SDRecordMarker commentStamp: 'rjl 4/10/2008 21:32' prior: 0!
I exist to mark a reference to another object in the serialized file, I'm replaced by a real object when I'm deserialized.  I'm basically a manually invoked proxy being used because getting a real proxy that inherits from ProtoObject to serialize seemed hurculean.!

!SDRecordMarker class methodsFor: 'testing' stamp: 'sas 4/10/2008 10:34'!
deservesSandstoneReferenceMark
	"Answers true if the receiver deserves 
	to be marked (or sub marked) as Sandstone reference"

	^ false! !

!SDRecordMarker methodsFor: 'converting' stamp: 'rjl 3/10/2008 21:00'!
asOriginalObject
	"this needs to call cachedLoadFrom not findById in order to allow cycles in the serialized graph
	this allows an object that references an unloaded object during the loadAll to force it's load instead of failing"
	^ (Smalltalk at: className) cachedLoadFrom: id! !

!SDRecordMarker methodsFor: 'accessing' stamp: 'rjl 12/11/2007 09:17'!
className
	^ className! !

!SDRecordMarker methodsFor: 'accessing' stamp: 'rjl 12/11/2007 09:17'!
className: anObject
	className := anObject! !

!SDRecordMarker methodsFor: 'accessing' stamp: 'rjl 12/11/2007 09:17'!
id
	^ id! !

!SDRecordMarker methodsFor: 'accessing' stamp: 'rjl 12/11/2007 09:17'!
id: anObject
	id := anObject! !

!UndefinedObject class methodsFor: '*sandstonedb-testing' stamp: 'sas 4/9/2008 13:31'!
deservesSandstoneReferenceMark
	"Answers true if the receiver deserves 
	to be marked (or sub marked) as Sandstone reference"
	
	^ false! !

!Dictionary methodsFor: '*sandstonedb' stamp: 'rjl 12/30/2007 23:32'!
replaceAll: anObject with: anotherObject
	self at: (self keyAtValue: anObject) put: anotherObject! !

!LookupKey class methodsFor: '*sandstonedb-as yet unclassified' stamp: 'sas 4/9/2008 13:43'!
deservesSandstoneReferenceMark
	"Answers true if the receiver deserves 
	to be marked (or sub marked) as Sandstone reference"
	
	^ true! !

Error subclass: #SDError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SandstoneDb'!

!SDError commentStamp: 'rjl 4/10/2008 21:36' prior: 0!
I'm just an abstract error!

SDError subclass: #SDCommitError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SandstoneDb'!

!SDCommitError commentStamp: 'rjl 4/10/2008 21:33' prior: 0!
Clearly, I'm thrown on a commit failure, duh!!'!

SDError subclass: #SDLoadError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SandstoneDb'!

!SDLoadError commentStamp: 'rjl 4/10/2008 20:24' prior: 0!
I'm thrown when an object fails to load from disk, either because the file was corrupt or because it was deleted.!

!FFIConstants class methodsFor: '*sandstonedb-as yet unclassified' stamp: 'sas 4/9/2008 13:30'!
deservesSandstoneReferenceMark
	"Answers true if the receiver deserves 
	to be marked (or sub marked) as Sandstone reference"
	
	^ false! !

!Integer methodsFor: '*sandstonedb' stamp: 'rjl 4/10/2008 10:08'!
asReducedSumOfDigits
	"Answers the sum of the digits present in the 
	decimal representation of the receiver 
	but also repeating the procedure if the 
	answers is greater than 9."
	^ 10 <= self 
		ifFalse: [ self ]
		ifTrue: 
			[ self = 10 
				ifTrue: [ 0 ]
				ifFalse: 
					[ (self asString 
						inject: 0
						into: [ :sum :e | sum + e digitValue ]) asReducedSumOfDigits ] ]! !

!ExternalObject class methodsFor: '*sandstonedb-as yet unclassified' stamp: 'sas 4/9/2008 13:28'!
deservesSandstoneReferenceMark
	"Answers true if the receiver deserves 
	to be marked (or sub marked) as Sandstone reference"
	
	^ false! !

!Magnitude class methodsFor: '*sandstonedb-as yet unclassified' stamp: 'sas 4/9/2008 13:33'!
deservesSandstoneReferenceMark
	"Answers true if the receiver deserves 
	to be marked (or sub marked) as Sandstone reference"
	
	^ false! !

!Set methodsFor: '*sandstonedb' stamp: 'rjl 12/30/2007 23:35'!
replaceAll: anObject with: anotherObject 
	self
		remove: anObject;
		add: anotherObject! !

!Collection class methodsFor: '*sandstonedb' stamp: 'rjl 4/10/2008 20:22'!
sandstoneReferenceMarkerSupport
	"Answers a reference marker support which 
	is specialized in the instropection of the receiver"
	^ SDCollectionReferenceMarker! !

!Collection class methodsFor: '*sandstonedb' stamp: 'rjl 4/10/2008 21:07'!
sandstoneReferenceResolverSupport
	"Answers a support to reify the receiver
	references to sandstone active records"
	^ SDCollectionReferenceResolver! !

!Collection methodsFor: '*sandstonedb' stamp: 'rjl 12/30/2007 23:22'!
replaceAll: anObject with: anotherObject
	self subclassResponsibility! !

!Array class methodsFor: '*sandstonedb-as yet unclassified' stamp: 'sas 4/9/2008 13:54'!
deservesSandstoneReferenceMark
	"Answers true if the receiver deserves 
	to be marked (or sub marked) as Sandstone reference"
	
	^ {Array.
	DependentsArray.
	WeakArray} includes: self! !

!InstructionStream class methodsFor: '*sandstonedb-as yet unclassified' stamp: 'sas 4/9/2008 13:27'!
deservesSandstoneReferenceMark
	"Answers true if the receiver deserves 
	to be marked (or sub marked) as Sandstone reference"
	
	^ false! !

!SmartRefStream methodsFor: '*sandstonedb' stamp: 'rjl 4/10/2008 21:39'!
sandstoneNextAndClose
	"Speedy way to grab one object. Only use when we are inside an object
	binary file.I created my own version of nextAndClose to get rid of the 
	interactive stuff the real one tries to do.  If I fail, then I fail, so be it"
	| obj |
	byteStream peek = ReferenceStream versionCode 
		ifFalse: [ SDLoadError signal: 'old object file' ].
	obj := self next.
	self close.
	^ obj! !

!SmartRefStream methodsFor: '*sandstonedb' stamp: 'rjl 3/31/2008 21:34'!
sandstoneNextPutObjOnly: anObject 
	"Really write three objects: (version, class structure, object).  But only when called from the outside.  Not in fileOut format.  No class definitions will be written for instance-specific classes.  Error if find one.  (Use nextPut: instead)"
	| info |
	topCall 
		ifNil: 
			[ topCall := anObject.
			super nextPut: ReferenceStream versionCode.
			info := self instVarInfo: anObject.
			self uniClasesDo: [ :cls | cls error: 'Class defn not written out.  Proceed?' ].
			objCount := 0.
			super nextPut: info.
			super nextPut: anObject.	"<- the real writing"
			"Class inst vars not written here!!"
			topCall := nil
			"references is an IDict of every object that got written
			(in case you want to take statistics)"
			"Transcript cr; show: structures keys printString."	"debug"	"reset it" ]
		ifNotNil: [ super nextPut: anObject ]! !

!Boolean class methodsFor: '*sandstonedb-as yet unclassified' stamp: 'sas 4/9/2008 13:27'!
deservesSandstoneReferenceMark
	"Answers true if the receiver deserves 
	to be marked (or sub marked) as Sandstone reference"
	
	^ false! !

!Morph class methodsFor: '*sandstonedb-as yet unclassified' stamp: 'sas 4/9/2008 13:31'!
deservesSandstoneReferenceMark
	"Answers true if the receiver deserves 
	to be marked (or sub marked) as Sandstone reference"
	
	^ false! !
